package message;

import java.io.*;

/**
* <p><code>Message</code> represents one of the messages which are exchanged between the federate and the RTIA.
* It contains the information that are used in all messages (the <code>Type</code> of the message, its <code>Exception</code> and its <code>FederationTime</code>)</p>
*
* @author <a href = "mailto:yannick.bisiaux@supaero.fr">Yannick Bisiaux</a>, <a href = "mailto:ronan.bossard@supaero.fr">Ronan Bossard</a>, <a href = "mailto:samuel.reese@supaero.fr">Samuel Reese</a>
* @version 1.0
*/

@SuppressWarnings("serial")
public class Message implements Serializable {

	private Type type;
	private Exception exception;
	private FederationTime fed_time;
	
	/**Creates a new instance of <code>Message</code>. The default initialization set Type to NOT_USED, Exception to e_NO_EXCEPTION, FederationTime to 0.
	 * 
	 */
	public Message(){
		this.type = Type.NOT_USED;
		this.exception = Exception.e_NO_EXCEPTION;
		this.fed_time = new FederationTime(0);
	}
	
	/**Creates a new instance of <code>Message</code>.
	 * 
	 * @param type the <code>Type</code> to set
	 */
	public Message(Type type_){
		this.type = type_;
		this.exception = Exception.e_NO_EXCEPTION;
		this.fed_time = new FederationTime(0);
	}
	
	/**Creates a new instance of <code>Message</code>.
	 * 
	 * @param type the <code>Type</code> to set
	 * @param exception the <code>Exception</code> to set
	 * @param fed_time the <code>FederationTime</code> to set
	 */
	public Message(Type type, Exception exception, FederationTime fed_time) {
		this.type = type;
		this.exception = exception;
		this.fed_time = fed_time;
	}

	/**Writes in the <code>MessageBuffer</code> the state of the attributes of the <code>Message</code>.
	 * The order will be the <code>Type</code>, the <code>Exception</code> and the federation time of the <code>Message</code>.
	 * 
	 * @param msgBuffer the <code>MessageBuffer</code> in which the method will write the state of the <code>Message</code>
	 */
	public void writeObject(MessageBuffer msgBuffer){
		msgBuffer.write(this.type.getType(), "type");
		msgBuffer.write(this.exception.getException(), "exception");
		msgBuffer.write(this.fed_time.getFed_time(), "fed_time");
	}
		
	/**Gives the correct values to the attributes by reading a <code>MessageBuffer</code> in which are stored those values.
	 * 
	 * @param msgBuffer the <code>MessageBuffer</code> from which the method will read the state of the <code>Message</code>
	 */
	public void readObject(MessageBuffer msgBuffer){
		//this.type = Type.reverseType.get(msgBuffer.readInt());
		this.exception = Exception.reverse.get(msgBuffer.readInt());
		this.fed_time.setFed_time(msgBuffer.readDouble());
	}

	/**<p>Creates a new default instance of a specialized <code>Message</code>.
	 * The kind of specialized class <code>Message</code> created is linked with the number in input.</p>
	 * 
	 * <p>For instance, <code>Message.virtualCreate(1)</code> will return a default instance of <code>MSG_Create_Federation_Execution</code>.</p>
	 * 
	 * @param n an <code>integer</code> which represents the <code>Type</code> (kind) of <code>Message</code> you want to create
	 * @return a default instance of the specialized <code>Message</code>
	 */
	public static Message virtualCreate(int n){
		String message_class;
		
		message_class = conversionType(n);
		
		try {
			Class c = Class.forName("message.MSG_" + message_class);//We have to add "message.MSG_" to have the package and the suffix of the class name.
			try {
				Object obj = c.newInstance();
				Message msg = (Message)obj;
				msg.setType(Type.reverseType.get(n));
				return msg;
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InstantiationException e) {
				e.printStackTrace();
			}
		} catch(ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	//This function takes in input the number of a Type of Message. It returns the name of the class we want to call.
	//Ex : conversionType(2) -> Destroy_Federation_Execution
	private static String conversionType(int n) {
		String type;
		String class_called;
		
		type = Type.reverseString.get(n);
		class_called = Message.correctClassName(type);
		
		return class_called;
	}
	
	//This function takes in input the string of a Type of Message. It returns the string with the capital letters in the right place (after each underscore).
	//Ex : correctClassName(DESTROY_FEDERATION_EXECUTION) -> Destroy_Federation_Execution
	private static String correctClassName(String str){
		StringBuffer buffer;
		boolean capitalizeNext = true; //The first letter will be a capital letter.
		char ch;
		
		str = str.toLowerCase(); //We begin by set all characters in lower case.
		buffer = new StringBuffer(str.length());
		
		for (int i = 0; i<str.length(); i++){
			ch = str.charAt(i);
			
			//If the character is an underscore, the next character will be a capital letter.
			if (ch == '_') {
                buffer.append(ch);
                capitalizeNext = true;
            } else if (capitalizeNext) {
                buffer.append(Character.toTitleCase(ch));
                capitalizeNext = false;
            } else {
                buffer.append(ch);
            }
		}
		return buffer.toString();
	}
	
	/**Returns the <code>Type</code> of this <code>Message</code>.
	 * @return the <code>Type</code>
	 */
	public Type getType() {
		return type;
	}

	/**Sets the <code>Type</code> of this <code>Message</code>.
	 * @param type_ the <code>Type</code> to set
	 */
	public void setType(Type type_) {
		this.type = type_;
	}

	/**Returns the <code>Exception</code> of this <code>Message</code>.
	 * @return the <code>Exception</code>
	 */
	public Exception getException() {
		return exception;
	}

	/**Sets the <code>Exception</code> of this <code>Message</code>.
	 * @param exception_ the <code>Exception</code> to set
	 */
	public void setException(Exception exception_) {
		this.exception = exception_;
	}

	/**Returns the <code>FederationTime</code> of this <code>Message</code>.
	 * @return the <code>FederationTime</code>
	 */
	public FederationTime getFed_time() {
		return fed_time;
	}

	/**Sets the <code>FederationTime</code> of this <code>Message</code>.
	 * @param fed_time_ the <code>FederationTime</code> to set
	 */
	public void setFed_time(FederationTime fed_time_) {
		this.fed_time = fed_time_;
	}
	
	/**This function is able to give a <code>String</code> which represents the state of the message.
	 * It gives its <code>Type</code>, its <code>Exception</code> and the federation time.
	 * @return the <code>String</code> which represents the object.
	 */
	public String toString(){
		String str;
		
		str = this.type + ", " + this.exception + ", federation time=" + this.fed_time.getFed_time();
		return str;
	}
}
